#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

Check the silviculture input tables for inconsistencies

    >>> from eu_cbm_hat.core.continent import continent
    >>> runner  = continent.combos['hat'].runners['AT'][-1]
    >>> runner.silv.check()

"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from plumbing.cache import property_cached

# Internal modules #

def keep_clfrs_without_question_marks(df, classifiers):
    """Check if there are questions mark in a classifier column
    and return a list of index columns that:

    :param (df) data frame of silviculture data
    :param (classifiers) list of classifier columns to check
    :output (list) list of classifiers that don't contain "?"

    The function performs the following:

    - keep a column if it contains no question marks

    - return an error if a column contains a mixture of question marks and
      other values

    - remove the column from output list if there are only question marks in
      that column

    Example use (will only work after simulation start):

        >>> from eu_cbm_hat.info.silviculture import keep_clfrs_without_question_marks
        >>> from eu_cbm_hat.core.continent import continent
        >>> runner  = continent.combos['hat'].runners['ZZ'][-1]
        >>> irw_frac = runner.silv.irw_frac.get_year(2016)
        >>> clfrs = list(runner.country.orig_data.classif_names.values())
        >>> keep_clfrs_without_question_marks(irw_frac, clfrs)

    """
    # TODO: The error raised when there are a mixture of other values and
    # question marks should be raised in the BaseSilvInfo.conv_clfrs() method.
    # Why is this not the case for irw_frac and the events_template?
    output_classifiers = []
    for classif_name in classifiers:
        values = df[classif_name].unique().tolist()
        if len(values) > 1 and "?" in values:
            msg =  "Mixture of question marks and other values"
            msg += f"not allowed in, column {classif_name}.\n"
            msg += f"The data frame contains the following columns:\n{df.columns}."
            raise ValueError(msg)
        # Remove classifiers that contain question marks only
        if "?" in values:
            continue
        # Add classifiers that don't contain question marks
        output_classifiers.append(classif_name)
    return output_classifiers


###############################################################################
class Silviculture:
    """
    Access to silvicultural information pertaining to a given country.
    This includes the following files:

    * irw_frac_by_dist.csv
    * vol_to_mass_coefs.csv
    * events_templates.csv
    * harvest_factors.csv

    Each is accessed by its own corresponding attribute:

    * `silv.irw_frac`
    * `silv.coefs`
    * `silv.events`
    * `silv.harvest`

    Loading this information will fail if you call `df` before a simulation is
    launched, because we need the internal SIT classifier and disturbance
    mapping. Before the simulation starts, it is only possible to load the raw
    versions of the input data as such: `silv.harvest.raw`.
    """

    def __init__(self, runner):
        # Default attributes #
        self.runner = runner
        # Shortcuts #
        self.country = self.runner.country

    @property_cached
    def irw_frac(self):
        return IRWFractions(self)

    @property_cached
    def coefs(self):
        return VolToMassCoefs(self)

    @property_cached
    def events(self):
        return EventsTemplates(self)

    @property_cached
    def harvest(self):
        return HarvestFactors(self)

    def check(self):
        """Check the consistency of silviculture input files"""
        self.events.check()
        self.coefs.check()
        self.harvest.check()
        self.irw_frac.check()

###############################################################################
class BaseSilvInfo:
    """
    A class that contains methods common to all silvicultural information
    files. You should inherit from this class.
    """

    def __init__(self, silv):
        # Default attributes #
        self.silv = silv
        # Shortcuts #
        self.runner  = self.silv.runner
        self.country = self.silv.country
        self.combo   = self.runner.combo
        self.code    = self.country.iso2_code

    #----------------------------- Properties --------------------------------#
    @property_cached
    def raw(self):
        """Data frame available in the input data"""
        return pandas.read_csv(self.csv_path,
                               dtype = {c:'str' for c in self.cols})

    @property_cached
    def cols(self):
        return list(self.country.orig_data.classif_names.values()) + \
               ['disturbance_type']

    @property_cached
    def dup_cols(self):
        return ['scenario'] + self.cols

    def check(self):
        """Perform various checks on the silviculture input tables"""
        # Make a check of duplicated entries #
        self.duplication_check()
        # Optional extra checks #
        if hasattr(self, 'extra_checks'): self.extra_checks()
        # Make a consistency check between dist_name and dist_id #
        if 'dist_type_name' in self.raw.columns:
            self.consistency_check()

    @property_cached
    def df(self):
        """Data frame with disturbance IDs and classifiers IDs converted to the
        internal IDs
        """
        self.check()
        # Load #
        df = self.raw.copy()
        # Drop the names which are useless #
        if 'dist_type_name' in self.raw.columns:
            df = df.drop(columns='dist_type_name')
        # Convert the disturbance IDs to the real internal IDs #
        df = self.conv_dists(df)
        # Convert the classifier IDs to the real internal IDs #
        df = self.conv_clfrs(df)
        # Return #
        return df

    #------------------------------- Methods ---------------------------------#
    def conv_dists(self, df):
        """
        Convert the disturbance IDs such as `20` and `22` into their
        internal simulation ID numbers that are defined by SIT.
        """
        # Get the conversion mapping and invert it #
        id_to_id = self.runner.simulation.sit.disturbance_id_map
        id_to_id = {v:k for k,v in id_to_id.items()}
        # Check that all IDs can be converted to an internal ID
        cannot_convert = df['disturbance_type'].dropna().map(id_to_id).isna()
        if any(cannot_convert):
            msg = f"In the file {self.csv_path}, the disturbance type(s) "
            msg += f"{df['disturbance_type'][cannot_convert].unique()} "
            msg += "cannot be converted to an internal disturbance id, using the "
            msg += "following mapping dictionary."
            raise ValueError(msg, id_to_id)
        # Apply the mapping to the dataframe #
        df['disturbance_type'] = df['disturbance_type'].map(id_to_id)
        # Return #
        return df

    def conv_clfrs(self, df):
        """
        Convert the classifier values such as `PA` and `QA` into their
        internal simulation ID numbers that are defined by SIT.
        """
        # Get all the conversion mappings, for each classifier #
        all_maps = self.runner.simulation.sit.classifier_value_ids.items()

        # Apply each of them to the dataframe #
        for classif_name, str_to_id in all_maps:
            if classif_name not in df.columns:
                continue

            # Handle question marks
            values = df[classif_name].unique()
            if len(values) == 1 and values[0] == "?":
                continue
            if len(values) > 1 and "?" in values:
                msg =  "Mixture of question marks and other values"
                msg += f"not allowed in %s, column {classif_name}"
                raise ValueError(msg % self)

            # Convert classifiers to their internal id
            df[classif_name] = df[classif_name].map(str_to_id)

        # Return #
        return df

    def consistency_check(self):
        # Get mapping dictionary from ID to full description #
        id_to_name = self.country.orig_data['disturbance_types']
        id_to_name = dict(zip(id_to_name['dist_type_name'],
                              id_to_name['dist_desc_input']))
        # Compare #
        names = self.raw['disturbance_type'].map(id_to_name)
        orig  = self.raw['dist_type_name']
        comp  = orig == names
        # Raise exception #
        if not all(comp):
            msg = "Names don't match IDs in '%s'.\n" % self.csv_path
            msg += "Names derived from the IDs:\n"
            msg += str(names[~comp])
            msg += "\n\nNames in the user file:\n"
            msg += str(orig[~comp])
            raise Exception(msg)

    def duplication_check(self):
        # What columns are we going to check duplication on #
        cols = self.dup_cols
        # Get duplicated rows #
        dups = self.raw.duplicated(subset=cols, keep=False)
        # Assert #
        if any(dups):
            msg = "There are duplicated entries in the file '%s'."
            msg += "\nThe duplicated rows are shown below:\n\n"
            msg += str(self.raw.loc[dups, cols])
            raise Exception(msg % self.csv_path)

    def get_year(self, year):
        # Case number 1: there is only a single scenario specified #
        if isinstance(self.choices, str): scenario = self.choices
        # Case number 2: the scenarios picked vary according to the year #
        else: scenario = self.choices[year]
        # Retrieve by query #
        df = self.df.query("scenario == '%s'" % scenario)
        # Drop the scenario column #
        df = df.drop(columns='scenario')
        # Check there is data left #
        assert not df.empty
        # Return #
        return df

###############################################################################
class IRWFractions(BaseSilvInfo):
    """
    Gives access the industrial roundwood fractions, per disturbance type,
    for the current simulation run.
    """

    @property
    def choices(self):
        """Choices made for `irw` fraction in the current combo."""
        return self.combo.config['irw_frac_by_dist']

    @property
    def csv_path(self):
        return self.country.orig_data.paths.irw_csv

###############################################################################
class VolToMassCoefs(BaseSilvInfo):
    """
    Gives access to the coefficients that enable the conversion from
    wood volume to wood mass.
    """

    @property
    def csv_path(self):
        return self.country.orig_data.paths.vol_to_mass_csv

    @property
    def cols(self):
        return ['forest_type']

    @property
    def dup_cols(self):
        return self.cols

    @property_cached
    def df(self):
        # Make a check of duplicated entries #
        self.duplication_check()
        # Load #
        df = self.raw.copy()
        # Convert the classifier IDs to the real internal IDs #
        df = self.conv_clfrs(df)
        # Return #
        return df

###############################################################################
class EventsTemplates(BaseSilvInfo):
    """
    Gives access to the dynamic events that have to be generated to
    satisfy the harvest.
    """

    @property
    def choices(self):
        """Choices made for the events templates in the current combo."""
        return self.combo.config['events_templates']

    @property
    def csv_path(self):
        return self.country.orig_data.paths.events_templates

    @property
    def dup_cols(self):
        return list(self.country.orig_data.classif_names.values()) + \
               ['scenario', 'sw_start', 'sw_end', 'hw_start', 'hw_end'] + \
               ["last_dist_id"]

    def extra_checks(self):
        # Guarantee no difference between sw_start and hw_start #
        assert all(self.raw['sw_start'] == self.raw['hw_start'])
        # Guarantee no difference between sw_end and hw_end #
        assert all(self.raw['sw_end'] == self.raw['hw_end'])
        # Guarantee we don't use max_since_last_dist #
        assert all(self.raw['max_since_last_dist'] == -1)

###############################################################################
class HarvestFactors(BaseSilvInfo):
    """
    Gives access to the data in the file `harvest_factors.csv`.
    """

    @property
    def choices(self):
        """Choices made for the harvest factors in the current combo."""
        return self.combo.config['harvest_factors']

    @property
    def csv_path(self):
        return self.country.orig_data.paths.harvest_factors

    @property
    def cols(self):
        cols = ['forest_type', 'mgmt_type', 'disturbance_type', 'con_broad']
        cols = cols + ['product_created', 'silv_practice']
        return cols

    @property
    def join_cols(self):
        """Keep only columns that are not empty as join columns"""
        join_cols = []
        for col in self.cols:
            if not any(self.df[col].isna()):
                join_cols.append(col)
        return join_cols

    def extra_checks(self):
        """Check the raw data for empty columns, proportions that sum to one.

        Check the raw data frame so it an be checked at the beginning of the simulation.
        """
        index = ["scenario", "product_created"]
        cols = list(set(self.cols) - set(["product_created"]))
        df_check = self.raw.groupby(index)[cols].agg(lambda x: len(x.isna().unique()))
        for col in cols:
            if any(df_check[col]>1):
                df_wrong = df_check[df_check[col]>1]
                msg = "For a given scenario and a given product, "
                msg += "A join column can either be completely empty or full, "
                msg += "but it cannot be incomplete i.e. "
                msg += "with some missing values and some values."
                msg += f"Check column: {col} in scenario {df_wrong.index.to_list()}"
                raise ValueError(msg)
        # Check that the skew factors sum to one by scenario and product group
        df_long = self.raw.melt(id_vars = self.cols + ["scenario"])
        index = ["scenario", "product_created", "variable"]
        df_long["value_sum"] = df_long.groupby(index)["value"].transform(sum)
        df_long_irw = df_long.query("product_created=='irw_and_fw'")
        if not all(df_long_irw["value_sum"] == 1):
            msg = "The following skew factors do not sum to one"
            raise ValueError(msg, df_long_irw.query("value_sum !=1"))



"""
Test QAQC methods

Execute the test suite from bash with py.test as follows:

    cd ~/repos/eu_cbm_hat/eu_cbm_hat
    pytest

"""

import numpy as np
import pandas
import pytest
from eu_cbm_hat.core.continent import continent
combo  = continent.combos['hat']
runner = combo.runners['ZZ'][-1]





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #

# First party modules #
from plumbing.common import camel_to_snake

# Internal modules #
from eu_cbm_hat.pump.classifiers import make_classif_df

###############################################################################
class InternalData(object):
    """
    This class will provide access to the data of a simulation as it is
    running as several pandas data frames.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent
        self.sim    = self.runner.simulation

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__, self.runner.short_name)

    #--------------------------- Special Methods -----------------------------#
    def __getitem__(self, item):
        """Read a dataframe from the `results` attribute."""
        # Special case for values #
        if item == 'values': return self.sim.sit.classifier_value_ids
        # Load #
        df = getattr(self.sim.results, item).copy()
        # Modify column names #
        df.columns = df.columns.to_series().apply(camel_to_snake)
        # Rename column names #
        df = df.rename(columns = {'input': 'area'})
        # Return #
        return df

    #----------------------------- Properties --------------------------------#
    @property
    def classif_df(self):
        return make_classif_df(self['values'], self['classifiers'])

    #------------------------------- Methods ---------------------------------#
    def load(self, name, with_clfrs=True, add_year=True):
        """
        Loads one of the files that was previously saved from the `libcbm_py`
        simulation and adds information to it.
        """
        # Load from CSV #
        df = self[name]
        # Optionally join classifiers #
        if with_clfrs:
            df = df.merge(self.classif_df, 'left', ['identifier', 'timestep'])
        # Add year if there is a timestep column #
        if add_year:
            if 'timestep' in df.columns:
                df['year'] = self.runner.country.timestep_to_year(df['timestep'])
                df = df.drop(columns='timestep')
        # Add age class information to merge with inventory #
        if 'age' in df.columns:
            df['age_class'] = df.age // 10 + 1
            df['age_class'] = 'AGEID' + df.age_class.astype(str)
        # Return #
        return df



"""Copy package internal test data to the eu_cbm_data folder

Usage:

    from eu_cbm_hat.tests.copy_data import copy_test_data
    copy_test_data()

"""
import shutil
from pathlib import Path
from eu_cbm_hat import module_dir, eu_cbm_data_dir

def copy_test_data():
    """Copy tests data from the package internal test folder
    to the eu_cbm_data folder"""
    orig_path = Path(module_dir) / "tests/eu_cbm_data"
    dest_path = Path(eu_cbm_data_dir)
    # Create the data folder if it doesn't exist
    # dest_path.mkdir(exist_ok=True, parents=True)
    # Copy ZZ test data to the eu_cbm_data directory
    # msg = f"\nIf the {dest_path} contains data already, "
    # msg += "this command will erase and replace the data :\n - "
    # if input(msg + "\nPlease confirm [y/n]:") != "y":
    #     print("Cancelled.")
    # else:
    shutil.copytree(orig_path, dest_path, dirs_exist_ok=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #

# Third party modules #
import pandas

# First party modules #
from plumbing.cache import property_cached
from plumbing.common import camel_to_snake

# Internal modules #

###############################################################################
class Fluxes:
    """
    Access to the information about fluxes that each disturbance creates.
    This data is extracted from the AIDB.

    Example use check that clear cut disturbances are the ones that remove most of the merch biomass:

        >>> from eu_cbm_hat.core.continent import continent
        >>> runner = continent.combos['hat'].runners['ZZ'][-1]
        >>> # Run at least one present time step so that SIT is available
        >>> runner.num_timesteps = runner.country.base_year - runner.country.inventory_start_year + 1
        >>> runner.run()
        >>> # Rank fluxes to merchantable pool to check clear cut disturbances
        >>> # are actually the ones with the highest proportion in the dist matrix
        >>> cols = ['user_name', 'silv_practice', 'hardwood_merch_prod_prop', 'softwood_merch_prod_prop']
        >>> runner.fluxes.df.sort_values("hardwood_merch_prod_prop", ascending=False)[cols]

    """

    def __init__(self, runner, debug=False):
        # Default attributes #
        self.runner = runner
        # Shortcuts #
        self.country = self.runner.country
        self.aidb = self.country.aidb
        # Extra attributes #
        self.debug = debug

    @property_cached
    def df(self):
        # The link between the internal sit_id and the external user_id #
        sit_usr = self.runner.simulation.sit.disturbance_id_map
        df = pandas.DataFrame.from_dict(sit_usr, orient='index',
                                        columns=['user_id'])
        df = df.rename_axis('sit_id')
        df = df.reset_index()

        # Then the user_id is mapped to a user_desc #
        usr_desc = self.country.orig_data['disturbance_types']
        usr_desc = usr_desc.rename(columns={'dist_type_name':'user_id'})
        df = pandas.merge(df, usr_desc, on='user_id', how='left')
        df = df.rename(columns={'dist_desc_input':'user_name'})

        # The user_desc is mapped to an aidb_name #
        desc_name = self.country.associations.key_to_rows('MapDisturbanceType')
        desc_name = pandas.DataFrame.from_dict(desc_name, orient='index',
                                               columns=['aidb_name'])
        df = pandas.merge(df, desc_name, how='left',
                          left_on='user_name', right_index=True)

        # The aidb_name is mapped to an aidb_id #
        name_id = self.country.aidb.db.read_df('disturbance_type_tr')
        name_id = name_id.rename(columns={'disturbance_type_id': 'aidb_id'})
        name_id = name_id.set_index('name')['aidb_id']
        df = pandas.merge(df, name_id, how='left',
                          left_on='aidb_name', right_index=True)

        # The aidb_id is mapped to a flux matrix_id #
        table = 'disturbance_matrix_association'
        id_matrix = self.country.aidb.db.read_df(table)
        mapping = {'disturbance_type_id':   'aidb_id',
                   'disturbance_matrix_id': 'matrix_id'}
        id_matrix = id_matrix.rename(columns = mapping)
        id_matrix = id_matrix.groupby(['aidb_id'])
        id_matrix = id_matrix.agg({'matrix_id': 'unique'}).reset_index()
        assert all(id_matrix['matrix_id'].apply(len) == 1)
        id_matrix = id_matrix.explode('matrix_id')
        df = pandas.merge(df, id_matrix, on='aidb_id', how='left')

        # Get the ID numbers of every pool, especially `products` pool #
        pool_ids = self.country.aidb.db.read_df('pool')
        pool_ids = pool_ids.rename(columns={'id':   'pool_id',
                                            'code': 'pool_name'})
        prod_id = pool_ids.query('pool_name == "Products"')['pool_id'].item()

        # Take all matrices and keep only fluxes going to `products` pools #
        mtxs = self.country.aidb.db.read_df('disturbance_matrix_value')
        mtxs = mtxs.rename(columns={'disturbance_matrix_id': 'matrix_id'})
        mtxs = mtxs.query('sink_pool_id == %i' % prod_id)
        mtxs = mtxs.drop(columns='sink_pool_id')
        mtxs = mtxs.rename(columns={'source_pool_id': 'pool_id'})

        # Replace IDs by names of pools and pivot #
        mtxs = pandas.merge(mtxs, pool_ids, on='pool_id')
        mtxs = mtxs.drop(columns = 'pool_id')
        mtxs = mtxs.pivot(index   = 'matrix_id',
                          columns = 'pool_name',
                          values  = 'proportion')
        mtxs = mtxs.fillna(0)
        mtxs = mtxs.reset_index()

        # Optional debug message #
        if self.debug:
            msg = "Movements to the `products` pool are coming exclusively" \
                  " form the pools listed below (for the AIDB at '%s').\n\n"
            cols = list(mtxs.columns)
            cols.pop(cols.index('matrix_id'))
            msg += ', '.join(cols)
            print(msg % self.aidb.repo_file)

        # Add pool columns to df #
        df = pandas.merge(df, mtxs, on='matrix_id', how='left')
        df = df.fillna(0)

        # Modify column names #
        df.columns = df.columns.to_series().apply(camel_to_snake)
        from eu_cbm_hat.cbm.dynamic import DynamicSimulation
        sources   = DynamicSimulation.sources
        self.cols = [pool + '_prod_prop' for pool in sources]
        mapping   = dict(zip(sources, self.cols))
        df        = df.rename(columns = mapping)

        # We always want to have the eight source pool columns #
        for pool in self.cols:
            if pool not in df.columns: df[pool] = 0.0

        # The name we will be using for the join #
        df = df.rename(columns={'sit_id': 'disturbance_type'})

        # Return #
        return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Imports #
from setuptools import setup, find_namespace_packages
from os import path

# Load the contents of the README file #
this_dir = path.abspath(path.dirname(__file__))
readme_path = path.join(this_dir, 'README.md')
with open(readme_path, encoding='utf-8') as handle: readme = handle.read()

# Call setup #
setup(
    name             = 'eu_cbm_hat',
    version          = '0.6.4',
    description      = 'eu_cbm_hat is a python package for running carbon'
                       ' budget simulations.',
    license          = 'EUPL',
    url              = 'https://gitlab.com/bioeconomy/eu_cbm/eu_cbm_hat',
    author           = 'Lucas Sinclair',
    author_email     = 'lucas.sinclair@me.com',
    maintainer       = 'Paul Rougieux',
    packages         = find_namespace_packages(),
    install_requires = ['autopaths>=1.6.0', 'plumbing>=2.11.1',
                        'pymarktex>=1.4.6', 'pandas', 'simplejson', 'pyyaml',
                        'tqdm', 'p_tqdm', 'pyarrow'],
    extras_require   = {'extras': ['pystache', 'matplotlib', 'numexpr']},
    python_requires  = ">=3.8",
    scripts=['scripts/running/run_zz_in_temp_dir_without_eu_cbm_data.py'],
    long_description = readme,
    long_description_content_type = 'text/markdown',
    include_package_data = True,
)



"""
Test salvage logging methods

Execute the test suite from bash with py.test as follows:

    cd ~/repos/eu_cbm_hat/eu_cbm_hat
    py.test

Inspired by the use in pandas
https://github.com/pandas-dev/pandas/blob/main/pandas/tests/strings/test_find_replace.py
"""

from eu_cbm_hat.core.continent import continent

combo  = continent.combos['hat']
runner = combo.runners['ZZ'][-1]

# runner.num_timesteps = 30
# runner.run(keep_in_ram=True, verbose=True, interrupt_on_error=True)

def test_dist_4_followed_by_29():
    """Disturbance 4 followed by disturbance 29 salvage logging. Prepare input
    data for ZZ Trying to overwrite as much as possible of the input data with
    these data frames """
    # Overwrite the natural disturbance activity
    # events, inventory, growth curves and transitions


    # check the method "__call__" in "info/input_data.py" to see
    # who the data is copied from eu_cbm_data/countries/zz
    # to eu_cbm_data/output/hat/ZZ/0/input/csv before the model run
    # Some files are transformed from wide to long format in this process.

    # Overwrite the events template
    # reference	For	PA	LU00	H	E	35	con	1	Cur	irw_and_fw	1	FALSE	0	5	0	5
    # reference	For	PA	LU00	H	E	45	con	1	Cur	irw_and_fw	1	FALSE	0	5	0	5

    # Run the model


    # Check output
    print(runner.input_data["events"])

    # assert [1,2] == [1,2]
    # assert (np.array([1,2]) == np.array([1,2])).all()
    # np.testing.assert_allclose(np.array([1,2]),np.array([1,3]))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import re

# Third party modules #

# First party modules #

# Internal modules #
from eu_cbm_hat.qaqc.input_years import InputYears
from eu_cbm_hat.qaqc.silviculture import SilvCheck
from eu_cbm_hat.qaqc.aidb import AIDBCheck
from eu_cbm_hat.qaqc.expected_provided import ExpectedProvided


class Qaqc:
    """
    Quality Assurance and Quality Control methods attached to a runner
    """

    def __init__(self, runner):
        # Default attributes #
        self.runner = runner

    @property
    def input_years(self):
        """Check input years"""
        return InputYears(self)

    @property
    def silv_check(self):
        """Check the consistency of silviculture input files"""
        return SilvCheck(self)

    @property
    def expected_provided(self):
        """Check harvest expected versus provided"""
        return ExpectedProvided(self)

    @property
    def aidb_check(self):
        """Check the consistency of the aidb"""
        return AIDBCheck(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

This script will add the required flux aggregate indicators to the AIDB of
every country.

If you want to test the script on only one country use:

    >>> if adder.iso2_code != 'ZZ': continue
"""

# Built-in modules #
import os

# Third party modules #
from tqdm import tqdm

# First party modules #
from plumbing.cache import property_cached
from plumbing.timer import Timer

# Internal modules #
from eu_cbm_hat.core.continent import continent

###############################################################################
class AddIndicators(object):
    """
    This class will open each AIDB and modify three tables in it:

    * `flux_indicator`
    * `flux_indicator_sink`
    * `flux_indicator_source`

    The pool IDs:

        1   SoftwoodMerch
        2   SoftwoodFoliage
        3   SoftwoodOther
        4   SoftwoodCoarseRoots
        5   SoftwoodFineRoots
        6   HardwoodMerch
        7   HardwoodFoliage
        8   HardwoodOther
        9   HardwoodCoarseRoots
        10  HardwoodFineRoots
        18  SoftwoodStemSnag
        19  SoftwoodBranchSnag
        20  HardwoodStemSnag
        21  HardwoodBranchSnag
        26  Products
    """

    def __repr__(self):
        return '%s object code "%s"' % (self.__class__,
                                        self.iso2_code)

    def __init__(self, country):
        # Main attributes #
        self.country = country
        # Shortcuts #
        self.iso2_code = self.country.iso2_code
        self.db = self.country.aidb.db

    useless = ['DisturbanceSoftProduction',
               'DisturbanceHardProduction',
               'DisturbanceDOMProduction']

    new_ones = ['softwood_merch_to_product',
                'softwood_other_to_product',
                'softwood_stem_snag_to_product',
                'softwood_branch_snag_to_product',
                'hardwood_merch_to_product',
                'hardwood_other_to_product',
                'hardwood_stem_snag_to_product',
                'hardwood_branch_snag_to_product']

    nums    = [53, 54, 55, 56, 57, 58, 59, 60]

    sources = [1,   3, 18, 19,  6,  8, 20, 21]

    sink = 26

    def __call__(self, debug=False, remove_legacy=False):
        # Optional debugging messages #
        if debug:
            num_rows = self.db.count_entries('flux_indicator')
            print(self.iso2_code, num_rows)
        # Read tables #
        indics  = self.db.read_df("flux_indicator")
        sources = self.db.read_df("flux_indicator_source")
        sinks   = self.db.read_df("flux_indicator_sink")
        # Drop old useless indicators that were going to products #
        if remove_legacy:
            for indic_name in self.useless:
                # Get the ID #
                num = indics.query("name == '%s'" % indic_name)['id'].item()
                # Drop from main table #
                indic_loc = indics.loc[indics['id'] == num].index
                indics = indics.drop(indic_loc)
                # Drop from secondary tables #
                loc = sources.loc[sources['flux_indicator_id'] == num].index
                sources = sources.drop(loc)
                loc = sinks.loc[sinks['flux_indicator_id'] == num].index
                sinks = sinks.drop(loc)
        # Add our eight new indicators #
        for num, new, source in zip(self.nums, self.new_ones, self.sources):
            # Add to the main table #
            indics = indics.append({'id':              num,
                                    'name':            new,
                                    'flux_process_id': 3 },
                                    ignore_index = True)
            # Add to the source table #
            highest = sources['id'].max()
            sources = sources.append({'id':                highest + 1,
                                      'flux_indicator_id': num,
                                      'pool_id':           source},
                                      ignore_index = True)
            # Add to the sink table #
            highest = sinks['id'].max()
            sinks = sinks.append({'id':                highest + 1,
                                  'flux_indicator_id': num,
                                  'pool_id':           self.sink},
                                  ignore_index = True)
        # Write tables #
        self.db.write_df(indics,  "flux_indicator")
        self.db.write_df(sources, "flux_indicator_source")
        self.db.write_df(sinks ,  "flux_indicator_sink")

###############################################################################
if __name__ == '__main__':
    # Make adder objects, one per country #
    adders = [AddIndicators(c) for c in continent]
    # Message #
    print("Adding flux indicators.")
    print("-------------")
    # Print timer start #
    timer = Timer()
    timer.print_start()
    # Run them all #
    for adder in tqdm(adders):
        adder()
    # Print end #
    timer.print_end()
    timer.print_total_elapsed()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Built-in modules #
import pickle

# Third party modules #
import pandas

# First party modules #
from autopaths.auto_paths import AutoPaths
from plumbing.cache import property_cached

# Internal modules #
from eu_cbm_hat.info.internal_data import InternalData

###############################################################################
class OutputData(InternalData):
    """
    This class will provide access to the output data of a Runner
    as several pandas data frames.

        >>> print(runner.output.load('pools'))
        >>> print(runner.output.load('flux'))
    """

    # The file "results.parquet" is called "results" because
    # calling it flux_pool.parquet would return an auto_paths error:
    #   Exception: Found several paths matching 'flux'
    # We want to avoid this error so that existing notebooks remain compatible.
    all_paths = """
    /output/csv/
    /output/csv/values.pickle
    /output/csv/area.csv.gz
    /output/csv/classifiers.csv.gz
    /output/csv/flux.csv.gz
    /output/csv/parameters.csv.gz
    /output/csv/pools.csv.gz
    /output/csv/state.csv.gz
    /output/csv/extras.csv.gz
    /output/csv/events.csv.gz
    /output/csv/results.parquet
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.runner = parent
        self.sim    = self.runner.simulation
        # Directories #
        self.paths = AutoPaths(self.parent.data_dir, self.all_paths)

    #----------------------------- Properties --------------------------------#
    @property_cached
    def extras(self):
        """
        This is a dataframe that will contain custom reporting information
        that is filled in by the `dynamics_fun` of a running simulation.
        It has one row for each year of the simulation run and contains
        information about harvest volumes.
        """
        return pandas.DataFrame()

    @property_cached
    def events(self):
        """
        This is a dataframe that will contain custom reporting information
        that is filled in by the `dynamics_fun` of a running simulation.
        It contains the dynamic events generated by H.A.T. for every timestep
        of a simulation.
        """
        return pandas.DataFrame()

    #--------------------------- Special Methods -----------------------------#
    def __getitem__(self, item):
        """Read any CSV or pickle file with the passed name."""
        # Find the path #
        path = self.paths[item]
        # If it is a CSV #
        if '.csv' in path.name:
            return pandas.read_csv(str(path), compression='gzip')
        # If it is a python pickle file #
        with path.open('rb') as handle: return pickle.load(handle)

    def __setitem__(self, item, df):
        """
        Record a dataframe or python object to disk using the file with the
        passed name.
        """
        # Find the path #
        path = self.paths[item]
        # If it is a DataFrame #
        if isinstance(df, pandas.DataFrame):
            # Convert disturbance ids from the model internal id to the input id
            dist_map = self.sim.sit.disturbance_id_map
            dist_map[0] = "0" # Keep the non disturbance fluxes
            for dist_col in ["disturbance_type", "last_disturbance_type"]:
                if dist_col in df.columns:
                    df[dist_col] = df[dist_col].map(dist_map)
            return df.to_csv(str(path),
                             index        = False,
                             float_format = '%g',
                             compression  = 'gzip')
        # If it is a python object #
        with path.open('wb') as handle: return pickle.dump(df, handle)

    #------------------------------- Methods ---------------------------------#
    def save(self, verbose=False):
        """
        Save all the information of interest from the simulation to disk before
        the whole cbm object is removed from memory.
        """
        # Message #
        self.parent.log.info("Saving final simulations results to disk.")
        # The classifier values #
        self['values']      = self.sim.sit.classifier_value_ids
        # All the tables that are within the SimpleNamespace of `sim.results` #
        tables = ['area', 'classifiers', 'flux', 'parameters',
                  'pools', 'state']
        # Loop and save them #
        for table in tables:
            if verbose:
                self.parent.log.info("Writing and compressing `%s`." % table)
            self[table] = self.runner.internal[table]
            if verbose:
                self.parent.timer.print_elapsed()
        # Save extra information to csv files using the above __setitem__() method
        self['extras']      = self.extras.reset_index()
        self['events']      = self.events
        # Merge all pools and fluxes and save them to a parquet file
        result = self.runner.internal
        classifiers = self.classif_df
        classifiers["year"] = self.runner.country.timestep_to_year(classifiers["timestep"])
        index = ['identifier', 'timestep']
        df = (result['parameters']
              .merge(result['flux'], 'left', on = index)
              .merge(result['state'], 'left', on = index)
              .merge(result['pools'], 'left', on = index)
              .merge(classifiers, 'left', on = index)
             )
        # Keep the internal id for debugging purposes
        df["disturbance_type_internal"] = df["disturbance_type"]
        # Convert the disturbance IDs from their internal simulation IDs that
        # are defined by SIT into the user defined equivalent string.
        id_to_id = self.runner.simulation.sit.disturbance_id_map
        df["disturbance_type"] = df["disturbance_type"].map(id_to_id)
        df["disturbance_type"] = df["disturbance_type"].astype(int)
        # Add age class information
        df['age_class'] = df.age // 10 + 1
        df['age_class'] = 'AGEID' + df.age_class.astype(str)
        # Write to a parquet file
        df.to_parquet(self.paths["results"])
        # Timer #
        self.parent.timer.print_elapsed()

    @property_cached
    def pool_flux(self):
        """Load the main result data frame where the following tables have been
        merged: area, params, flux, state, pools

        Example usage:

            from eu_cbm_hat.core.continent import continent
            runner = continent.combos['hat'].runners['ZZ'][-1]
            pool_flux = runner.output.pool_flux

        This loads faster than the following equivalent code

            params = runner.output.load('parameters', with_clfrs=False)
            flux = runner.output.load('flux', with_clfrs=False)
            state = runner.output.load('state', with_clfrs=False)
            pools = runner.output.load('pools', with_clfrs=False)
            classifiers = runner.output.classif_df
            classifiers["year"] = runner.output.runner.country.timestep_to_year(classifiers["timestep"])
            index = ['identifier', 'year']
            df = (params
                  .merge(flux, 'left', on = index)
                  .merge(state, 'left', on = index)
                  .merge(classifiers, 'left', on = index)
                  .merge(pools, 'left', on = index)
                 )
            df.equals(pool_flux)

        """
        return pandas.read_parquet(self.paths["results"])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Lucas Sinclair and Paul Rougieux.

JRC Biomass Project.
Unit D1 Bioeconomy.

This script will rename the header column of the file:

* /common/associations.csv

Before running this script the headers are simply "A", "B", "C".

After running this script, the new headers will be:

* "category"
* "name_input"
* "name_aidb"
"""

# Built-in modules #

# Third party modules #
import pandas
from tqdm import tqdm

# First party modules #

# Internal modules #
from eu_cbm_hat.core.continent import continent

###############################################################################
class RenameAssociations(object):

    def __init__(self, country):
        # Main attributes #
        self.country = country

    def __call__(self, verbose=False):
        # Get path #
        path = self.country.orig_data.paths.associations
        # Load dataframe #
        df = pandas.read_csv(str(path))
        # Modify dataframe #
        df.columns = ["category", "name_input", "name_aidb"]
        # Write dataframe back to disk #
        df.to_csv(str(path), index=False, float_format='%g')

###############################################################################
if __name__ == '__main__':
    # Make renamer objects, one per country #
    renamers = [RenameAssociations(c) for c in continent]
    # Run them all #
    for merger in tqdm(renamers):
        merger()
